#include <IRremote.h>
#include <SPI.h>
#include "Adafruit_FRAM_SPI.h"
// Pin Declarations
const int NUM_FUNCS = 6;
const int LEDS[NUM_FUNCS] = {A0, A1, A5, A2, A3, A4};
const int REC_LED = 4;

const int BUTTON_PIN = 6;
const int REC_SWITCH_PIN= 5;

const int RECV_PIN = 2;

const uint8_t FRAM_CS = 10;

// Code storage
IRrecv irrecv(RECV_PIN);
IRsend irsend;
Adafruit_FRAM_SPI fram = Adafruit_FRAM_SPI(FRAM_CS);

// Storage for the recorded code
struct CodeData {
  int type;
  unsigned long value;
  unsigned int raw[RAWBUF];
  int len;
  int toggle;
};
const int CODE_SIZE = sizeof(CodeData);

union CodeBuffer {
  CodeData data;
  byte bytes[CODE_SIZE];
};

CodeBuffer codes[NUM_FUNCS];

decode_results results;

enum State {idle, scrollForward, scrollReverse, hold, record};
//********** State Vars **********//
State state = idle;
int buttonState = LOW;
int lastButtonState = LOW;
unsigned long buttonTimeout = 0;
int recordState = LOW;
int activeFunction = 0;
unsigned long lastTime = 0;
const unsigned int scrollDelay = 1000;

//********** Setup **********//
void setup() {
  Serial.begin(9600);
  if (!fram.begin()){ 
    Serial.println("F");
    while(1);
  }
  irrecv.enableIRIn(); // Start the receiver
  pinMode(BUTTON_PIN, INPUT);
  pinMode(REC_SWITCH_PIN, INPUT);
  pinMode(REC_LED, OUTPUT);
  for (int i = 0; i < NUM_FUNCS; i++) {
    pinMode(LEDS[i], OUTPUT);
    framRead(getAddress(i), i);
  }
  digitalWrite(REC_LED, HIGH);
  delay(100);
  digitalWrite(REC_LED, LOW);
  Serial.println("EoS");
}

//********** IR Helpers **********//
void storeCode(decode_results *results, int func) {
  codes[func].data.type = results->decode_type;
  if (codes[func].data.type == UNKNOWN) {
    Serial.println("SR");
    codes[func].data.len = results->rawlen - 1;
    // To store raw codes:
    // Drop first value (gap)
    // Convert from ticks to microseconds
    // Tweak marks shorter, and spaces longer to cancel out IR receiver distortion
    for (int i = 1; i <= codes[func].data.len; i++) {
      if (i % 2) {
        // Mark
        codes[func].data.raw[i - 1] = (results->rawbuf[i]*USECPERTICK) - MARK_EXCESS;
        Serial.print(" m");
      } 
      else {
        // Space
        codes[func].data.raw[i - 1] = (results->rawbuf[i]*USECPERTICK) + MARK_EXCESS;
        Serial.print(" s");
      }
      Serial.print(codes[func].data.raw[i - 1], DEC);
    }
    Serial.println("");
  }
  else {
    if (codes[func].data.type == NEC) {
      //Serial.print("Received NEC: ");
      if (results->value == REPEAT) {
        // Don't record a NEC repeat value as that's useless.
        Serial.println("IGR");
        return;
      }
    } 
    else if (codes[func].data.type == SONY) {
      //Serial.print("Received SONY: ");
    } 
    else if (codes[func].data.type == PANASONIC) {
      //Serial.print("Received PANASONIC: ");
    }
    else if (codes[func].data.type == JVC) {
      //Serial.print("Received JVC: ");
    }
    else if (codes[func].data.type == RC5) {
      //Serial.print("Received RC5: ");
    } 
    else if (codes[func].data.type == RC6) {
      //Serial.print("Received RC6: ");
    }
    else if (codes[func].data.type == SAMSUNG) {
      //Serial.print("Received Samsung: ");
    }
    else {
      Serial.print("UCT");// Unexpected Code Type
      Serial.print(codes[func].data.type, DEC);
      Serial.println("");
    }
    Serial.println(results->value, HEX);
    codes[func].data.value = results->value;
    codes[func].data.len = results->bits;
  }
}

void sendCode(int repeat, int func) {
  if (codes[func].data.type == NEC) {
    if (repeat) {
      irsend.sendNEC(REPEAT, codes[func].data.len);
      //Serial.println("Sent NEC repeat");
    } 
    else {
      irsend.sendNEC(codes[func].data.value, codes[func].data.len);
      //Serial.print("Sent NEC ");
      Serial.println(codes[func].data.value, HEX);
    }
  } 
  else if (codes[func].data.type == SONY) {
    irsend.sendSony(codes[func].data.value, codes[func].data.len);
    //Serial.print("Sent Sony ");
    Serial.println(codes[func].data.value, HEX);
  } 
  else if (codes[func].data.type == PANASONIC) {
    irsend.sendPanasonic(codes[func].data.value, codes[func].data.len);
    //Serial.print("Sent Panasonic");
    Serial.println(codes[func].data.value, HEX);
  }
  else if (codes[func].data.type == JVC) {
    irsend.sendJVC(codes[func].data.value, codes[func].data.len, false);
    //Serial.print("Sent JVC");
    Serial.println(codes[func].data.value, HEX);
  }
  else if (codes[func].data.type == RC5 || codes[func].data.type == RC6) {
    if (!repeat) {
      // Flip the toggle bit for a new button press
      codes[func].data.toggle = 1 - codes[func].data.toggle;
    }
    // Put the toggle bit into the code to send
    codes[func].data.value = codes[func].data.value & ~(1 << (codes[func].data.len - 1));
    codes[func].data.value = codes[func].data.value | (codes[func].data.toggle << (codes[func].data.len - 1));
    if (codes[func].data.type == RC5) {
      //Serial.print("Sent RC5 ");
      Serial.println(codes[func].data.value, HEX);
      irsend.sendRC5(codes[func].data.value, codes[func].data.len);
    } 
    else {
      irsend.sendRC6(codes[func].data.value, codes[func].data.len);
      //Serial.print("Sent RC6 ");
      Serial.println(codes[func].data.value, HEX);
    }
  }
  else if (codes[func].data.type == SAMSUNG) {
    irsend.sendSAMSUNG(codes[func].data.value, codes[func].data.len);
    //Serial.print("Sent Samsung ");
    Serial.println(codes[func].data.value, HEX);
  }
  else if (codes[func].data.type == UNKNOWN /* i.e. raw */) {
    // Assume 38 KHz
    irsend.sendRaw(codes[func].data.raw, codes[func].data.len, 38);
    //Serial.println("Sent raw");
  }
}

//********** FRAM Helpers **********//
void framRead(uint16_t address, int index) {
  Serial.print("FR-R:");Serial.println(millis());
  for (int i = 0; i < CODE_SIZE; i++) {
    codes[index].bytes[i] = fram.read8(address++);
  }
  Serial.println(millis());
}

void framWrite(uint16_t address, int index) {
  Serial.println("FR-W");
  for (int i = 0; i < CODE_SIZE; i++) {
    fram.writeEnable(true);
    fram.write8(address++, codes[index].bytes[i]);
  }
  fram.writeEnable(false);
}

uint16_t getAddress(int index) {
return 1024 + 256 * index;
}

void framDump() {
  uint8_t value;
  for (uint16_t a = 1024; a < 3072; a++) {
    value = fram.read8(a);
    if ((a % 32) == 0) {
      Serial.print("\n 0x"); Serial.print(a, HEX); Serial.print(": ");
    }
    Serial.print("0x"); 
    if (value < 0x1) 
      Serial.print('0');
    Serial.print(value, HEX); Serial.print(" ");
  }
}

//********** Hardware Helpers **********//
void setLight(int ledIndex) {
  for (int i = 0; i < NUM_FUNCS; i++) {
    if (ledIndex == i) {
      digitalWrite(LEDS[i], HIGH);
    } else {
      digitalWrite(LEDS[i], LOW);
    }
  }
}

void checkButtonState() {
  int currentState = digitalRead(BUTTON_PIN);
  if (currentState == HIGH && buttonState == LOW) {
    onButtonPressed();
    buttonTimeout = millis();
    buttonState = currentState;
  } else if (millis() - buttonTimeout < 100) {
    Serial.println("TO");
  } else if (currentState == LOW && buttonState == HIGH) {
    onButtonReleased();
    buttonState = currentState;
  }
}

void checkRecordState() {
  int currentState = digitalRead(REC_SWITCH_PIN);
  if (currentState != recordState) {
    recordState = currentState;
    Serial.print("RS: "); Serial.println(recordState);
    digitalWrite(REC_LED, recordState);
  }
}

//********** Loop Helpers **********//
void scroll(int increment) {
  unsigned long currentTime = millis();
  if (currentTime - lastTime > scrollDelay) {
    lastTime = currentTime;
    int nextFunction = activeFunction + increment;
    if (nextFunction < 0) {
      nextFunction = 0;
      state = idle;
      setLight(-1);
    } else if (nextFunction > NUM_FUNCS - 1) {
      activeFunction = NUM_FUNCS - 1;
      state = scrollReverse;
    } else {
      activeFunction = nextFunction;
      setLight(activeFunction);
    }
  }
}

void recordSignal() {
  if (irrecv.decode(&results)) {
    Serial.print("Rec:");
    Serial.println(activeFunction);
    storeCode(&results, activeFunction);
    framWrite(getAddress(activeFunction), activeFunction);
    irrecv.resume(); // Resume receiver
    state = idle;
    setLight(-1);
  }
}

void transmitSignal(int repeat) {
  Serial.print("Tra:");
  Serial.print(activeFunction);
  Serial.print(" r");
  Serial.println(repeat);
  sendCode(repeat, activeFunction);
  delay(50);
}

//********** Click Helpers **********//
void startScroll() {
  state = scrollForward;
  lastTime = millis();
  setLight(activeFunction);
}

void onActionSelected() {
  if (recordState) {
    state = record;
  } else {
    state = hold;
  }
}

void onActionReleased() {
  setLight(-1);
  state = idle;
  irrecv.enableIRIn();
}

void cancelRecord() {
  setLight(-1);
  state = idle;
}

//********** Click Handlers **********//
void onButtonPressed() {
  Serial.println("Prs");
  switch (state) {
    case scrollForward: onActionSelected(); break;
    case scrollReverse: onActionSelected(); break;
    case record: cancelRecord(); break;
  }
}

void onButtonReleased() {
  Serial.println("Rel");
  switch (state) {
    case idle: startScroll(); break;
    case hold: onActionReleased(); break;
  }
}

//********** Loop **********//
void loop (){
  checkButtonState();
  checkRecordState();
  switch (state) {
    case scrollForward: scroll(1); break;
    case scrollReverse: scroll(-1); break;
    case record: recordSignal(); break;
    case hold: transmitSignal(buttonState == lastButtonState); break;
  }
  lastButtonState = buttonState;
}
